/**CFile***********************************************************************

  FileName    [bmcModel.c]

  PackageName [bmc]

  Synopsis    [Model Manager for the <tt>bmc</tt> package]

  Description []

  SeeAlso     []

  Author      [Alessandro Cimatti, Lorenzo Delana and Roberto Sebastiani]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static rbc_ptr bmc_mk_j_k_unrolling ARGS((Fsm_RbcPtr rbc_fsm, int j, int k));
static rbc_ptr bmc_mk_loopback      ARGS((Fsm_RbcPtr rbc_fsm, int k, int l));
static int bmc_succ_k_l_loop        ARGS((int time, int k, int l));
static rbc_ptr bmc_mk_fairness      ARGS((Fsm_RbcPtr rbc_fsm, int k, int l));
static rbc_ptr bmc_mk_fairness_aux  ARGS((node_ptr fairness, int k, int l));
static rbc_ptr bmc_mk_one_fairness  ARGS((rbc_ptr one_fairness, int k, int l));
static rbc_ptr bmc_mk_tableau_k_loops ARGS((Fsm_RbcPtr rbc_fsm, 
					    node_ptr ltl_wff, int k, int l));
static rbc_ptr 
bmc_mk_tableau_k_loops_depth1 ARGS((Fsm_RbcPtr rbc_fsm, node_ptr ltl_wff, 
				    int k));

static rbc_ptr bmc_mk_untiltableau_at_time_aux ARGS((node_ptr p, node_ptr q, 
						     int time, int k, int l, 
						     int steps));

static rbc_ptr bmc_mk_releasestableau_at_time_aux ARGS((node_ptr p, 
							node_ptr q, int time,
							int k, int l, 
							int steps));

static rbc_ptr bmc_mk_all_loops ARGS((Fsm_RbcPtr rbc_fsm, int k));


/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                        P A T H S   A N D   L O O P                        */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/**Function********************************************************************

  Synopsis           [Returns the path for the model from 0 to k, 
  taking into account initial conditions and invariants]

  Description        [Collects initial state and invariants at step zero as 
  bmc_mk_j_k_unrolling does not take into account of them]

  SideEffects        []

  SeeAlso            [bmc_mk_j_k_unrolling]

******************************************************************************/

rbc_ptr Bmc_Mk_k_Path (Fsm_RbcPtr rbc_fsm, int k)
{
  rbc_ptr Init = Compile_FsmRbcGetInit(rbc_fsm);
  rbc_ptr Invar = Compile_FsmRbcGetInvar(rbc_fsm);
  rbc_ptr Init_0 =  Bmc_CurrNextShiftAtTime(Init, 0);
  rbc_ptr Invar_0 = Bmc_CurrNextShiftAtTime(Invar, 0);
  rbc_ptr Path_0_k = bmc_mk_j_k_unrolling(rbc_fsm, 0, k);

  return ( Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), 
		      Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), Init_0, Invar_0), 
		      Path_0_k) );
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*           BOUNDED TABLEAU CONSTRUCTION ROUTINES FOR LTL FORMULAE          */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* 
 ****** WARNING ******
 Fairness handling is problematic.
 - for safety properties, we should not worry (check)
 - for liveness, we should
 
 Example for potential problem: counter modulo 255 with input that triggers
 transition requires at least 257 steps for fair cycle.

 If we ignore fairness conditions, we can endanger correctness of the whole
 system (bdd-based model checking always uses fairness).

*/


/**Function********************************************************************

  Synopsis           [Builds tableau without loop at time zero, taking account
  of fairnesses]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_MkTableau_k_Noloop(Fsm_RbcPtr rbc_fsm, node_ptr ltl_wff, int k)
{
  rbc_ptr Fairness_k = bmc_mk_fairness(rbc_fsm, k, BMC_NO_LOOP);
  
  /* lazy evaluation: */
  if ( Bmc_RbcIsFalse(Bmc_VarsMgrGetRbcMgr(), Fairness_k) ) {
    return Fairness_k;
  }
  else {
    rbc_ptr Tableau_k = Bmc_MkTableauAtTime(ltl_wff, 0, k, BMC_NO_LOOP);
    return(Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), Fairness_k, Tableau_k));
  }
}

/**Function********************************************************************

  Synopsis           [Builds the tableau at time zero. Loop is allowed]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_MkTableau_k_l(node_ptr ltl_wff, int k, int l)
{
  /* checks out the validity of [l, k] only if a loop exists: */
  nusmv_assert( (l == BMC_NO_LOOP) || (k >= l) );

  return(Bmc_MkTableauAtTime(ltl_wff, 0, k, l));
}


/**Function********************************************************************

  Synopsis           [Given a wff expressed in ltl builds the model-independent
  tableau at 'time' of a path formula bounded by \[k, l\]]

  Description        [This function is the entry point of a mutual recursive
  calling stack. All logical connectives are resolved, excepted for NOT, which
  closes the recursive calling stack. Variables and falsity/truth constants 
  close the recursion too]

  SideEffects        []

  SeeAlso            [Bmc_MkNextTableauAtTime, 
  Bmc_MkGloballyTableauAtTime, Bmc_MkEventuallyTableauAtTime, 
  Bmc_MkUntilTableauAtTime, Bmc_MkReleasesTableauAtTime]

******************************************************************************/
rbc_ptr Bmc_MkTableauAtTime(node_ptr ltl_wff, int time, int k, int l)
{
  rbc_ptr result=NULL;

  /* checks out the validity of [l, k] only if a loop exists: */
  nusmv_assert( (l == BMC_NO_LOOP) || (k >= l) );
  nusmv_assert(k >= time);

  switch (node_get_type(ltl_wff)) {
  case TRUEEXP:
    result = Bmc_RbcTruth(Bmc_VarsMgrGetRbcMgr());
    break;
    
  case FALSEEXP:
    result = Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());
    break;

  case DOT:
    Bmc_ExtendsMaxtime(time);
    result = Bmc_Statevar2Timedvar(ltl_wff, time);    
    break;

  case NOT:
    /* checks out that argument of NOT operator is actually a variable: */
    nusmv_assert( node_get_type(car(ltl_wff)) == DOT );
    Bmc_ExtendsMaxtime(time);
    result = Bmc_RbcNot(Bmc_VarsMgrGetRbcMgr(),
		      Bmc_Statevar2Timedvar(car(ltl_wff), time));
    break;

  case AND:
    result = Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), 
		   Bmc_MkTableauAtTime(car(ltl_wff), time, k, l),
		   Bmc_MkTableauAtTime(cdr(ltl_wff), time, k, l));
    break;

  case OR:
    result = Bmc_RbcOr(Bmc_VarsMgrGetRbcMgr(), 
		  Bmc_MkTableauAtTime(car(ltl_wff), time, k, l),
		  Bmc_MkTableauAtTime(cdr(ltl_wff), time, k, l));
    break;

  case IFF:
    result = Bmc_RbcIff(Bmc_VarsMgrGetRbcMgr(), 
		   Bmc_MkTableauAtTime(car(ltl_wff), time, k, l),
		   Bmc_MkTableauAtTime(cdr(ltl_wff), time, k, l));
    break;

  case OP_NEXT:
    result = Bmc_MkNextTableauAtTime(ltl_wff, time, k, l);
    break;

  case OP_GLOBAL:
    result = Bmc_MkGloballyTableauAtTime(car(ltl_wff), time, k, l);
    break;

  case OP_FUTURE: /* EVENTUALLY */
    result = Bmc_MkEventuallyTableauAtTime(car(ltl_wff), time, k, l);
    break; 

  case UNTIL:
    result = Bmc_MkUntilTableauAtTime(car(ltl_wff), cdr(ltl_wff), time, k, l);
    break; 

  case RELEASES:
    result = Bmc_MkReleasesTableauAtTime( car(ltl_wff), cdr(ltl_wff), time, 
					  k, l );
    break;

  case IMPLIES:
    internal_error("'Implies' should had been nnf-ed away!\n");

  case ATOM:
  case ARRAY:
  case EX:
  case AX:
  case EF:
  case AF:
  case EG:
  case AG:
  case EE:
  case AA:
  case EBF:
  case EBG:
  case ABF:
  case ABG:
  case BUNTIL:
  case MMIN:
  case MMAX:
  case APATH:
  case EPATH:
    internal_error( "Unexpected CTL operator, node type %d\n", 
		    node_get_type(ltl_wff) );

  default:
    /* no other type are available here: */
    nusmv_assert(false);
  }

  nusmv_assert(result != NULL); /*it must be assigned! */
  return result;
}

/**Function********************************************************************

  Synopsis           [Resolve the NEXT operator, building the tableau for
  its argument]

  Description        [Returns a falsity constants if the next operator leads 
  out of \[l, k\] and there is no loop]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr
Bmc_MkNextTableauAtTime(node_ptr ltl_wff, int time, int k, int l)
{
  int succtime;
  rbc_ptr tableau;
  
  nusmv_assert(node_get_type(ltl_wff) == OP_NEXT);
  nusmv_assert(time <= k);
  /* checks out the validity of [l, k] only if a loop exists: */
  nusmv_assert( (l == BMC_NO_LOOP) || (k >= l) );

  succtime = bmc_succ_k_l_loop(time, k, l);
 
  if (!(succtime == BMC_NO_LOOP)) {
    tableau = Bmc_MkTableauAtTime(car(ltl_wff), succtime, k, l);
  }
  else {
    tableau = Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());
  } 
  
  return(tableau);
}

/**Function********************************************************************

  Synopsis           [Resolves the future operator, and builds a conjunctive
  expression of tableaus, by iterating intime up to k in a different manner 
  depending on the \[l, k\] interval form]

  Description        [ltl_wff is the 'p' part in 'F p'. 
  If intime<=k is out of \[l, k\] or if there is no loop, 
  iterates from intime to k, otherwise iterates from l to k]
  
  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr
Bmc_MkEventuallyTableauAtTime(node_ptr ltl_wff, int intime, int k, int l)
{
  int time;
  rbc_ptr tableau;
  int stop_time; 

  nusmv_assert(intime <= k);
  /* checks out the validity of [l, k] only if a loop exists: */
  nusmv_assert( (l == BMC_NO_LOOP) || (k >= l) );

  tableau = Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());  

  /* there exist three cases: 
     1) no loop: iterates from k downto intime;
     2) loop, (intime <= l): iterates from k downto intime;
     3) loop, (l < intime <= k) : iterates from k downto l */
  if ( (l == BMC_NO_LOOP) || (intime <= l) ) {
    stop_time=intime;
  }
  else {
    /* the third case */
    stop_time = l; 
  }  

  for (time = k; time>=stop_time; --time) {

    /* lazy evaluation: */
    rbc_ptr tableau_at_time = Bmc_MkTableauAtTime(ltl_wff, time, k, l);
    if ( Bmc_RbcIsTrue(Bmc_VarsMgrGetRbcMgr(), tableau_at_time) ) {
      tableau = tableau_at_time;
      break;
    }
    tableau = Bmc_RbcOr( Bmc_VarsMgrGetRbcMgr(),  tableau_at_time, tableau );
  } /* loop */

  return(tableau); 
}

/**Function********************************************************************

  Synopsis           [As Bmc_MkEventuallyTableauAtTime, but builds a 
  conjunctioned expression in order to be able to assure a global constraint]

  Description        [ltl_wff is the 'p' part in 'G p']

  SideEffects        []

  SeeAlso            [Bmc_MkEventuallyTableauAtTime]

******************************************************************************/
rbc_ptr
Bmc_MkGloballyTableauAtTime(node_ptr ltl_wff, int intime, int k, int l)
{
  int time;
  rbc_ptr tableau;
  int stop_time;

  nusmv_assert(intime <= k);
  /* checks out the validity of [l, k] only if a loop exists: */
  nusmv_assert( (l == BMC_NO_LOOP) || (k >= l) );

  /* there exist three cases: 
     1) no loop: cannot assure nothing, so return falsity; 
     2) loop, (intime <= l): iterates from intime to k;
     3) loop, (l < intime <= k) : iterates from intime to k, and then from
        l to intime-1 (so more efficiently from l to k.)  */
  if (l == BMC_NO_LOOP) {
    tableau = Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());
  }
  else { 
    /* second and third cases */
    tableau = Bmc_RbcTruth(Bmc_VarsMgrGetRbcMgr());

    stop_time = min(intime, l);
    for (time=k; time >= stop_time; --time) {
      /* lazy evaluation: */
      rbc_ptr tableau_at_time = Bmc_MkTableauAtTime(ltl_wff, time, k, l);
      if ( Bmc_RbcIsFalse(Bmc_VarsMgrGetRbcMgr(), tableau_at_time) ) {
	tableau = tableau_at_time;
	break;
      }
      tableau = Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), tableau_at_time, tableau);
    } 
  }
  
  return(tableau); /* return */
}

/**Function********************************************************************

  Synopsis           [Builds an expression which evaluates the until operator]

  Description        [Carries out the steps number to be performed, depending 
  on l,k and time, then calls bmc_mk_untiltableau_at_time_aux]

  SideEffects        []

  SeeAlso            [bmc_mk_untiltableau_at_time_aux]

******************************************************************************/
rbc_ptr
Bmc_MkUntilTableauAtTime(node_ptr p, node_ptr q, int time, int k, int l)
{
  int steps;

  nusmv_assert (time <= k);

  if ( (l == BMC_NO_LOOP) || (time <= l) ) {
    steps = k + 1 - time; /* no loop, or time is external to ]l, k] */
  }
  else { 
    steps = k + 1 - l; /* loop, full round */
  }

  return(bmc_mk_untiltableau_at_time_aux(p, q, time, k, l, steps));
}


/**Function********************************************************************

  Synopsis           [Builds an expression which evaluates the release
  operator]

  Description        [Carries out the steps number to be performed, depending 
  on l,k and time, then calls bmc_mk_releasestableau_at_time_aux]

  SideEffects        []

  SeeAlso            [bmc_mk_releasestableau_at_time_aux]

******************************************************************************/
rbc_ptr
Bmc_MkReleasesTableauAtTime(node_ptr p, node_ptr q, int time, int k, int l)
{
  int steps;

  nusmv_assert (time <= k);

  if ( (l == BMC_NO_LOOP) || (time <= l) ) {
    steps = k + 1 - time; /* no loop, or time is external to ]l, k] */ 
  }
  else { 
    steps = k + 1 - l; /* loop, full round */
  }

  return(bmc_mk_releasestableau_at_time_aux(p, q, time, k, l, steps));
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                              TOP LEVEL FUNCTIONS                          */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/**Function********************************************************************

  Synopsis           [Returns the problem at length k with loopback l 
  (possibly BMC_NO_LOOP)]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/

rbc_ptr Bmc_MkProblem_k_l (Fsm_RbcPtr rbc_fsm, node_ptr ltl_wff, 
				   int k, int l)
{
  rbc_ptr Path_k = Bmc_Mk_k_Path(rbc_fsm, k);
  rbc_ptr Tableau_k_l = Bmc_MkTableau_k_l(ltl_wff, k, l);
  rbc_ptr Loopback_k_l = bmc_mk_loopback(rbc_fsm, k, l);
  rbc_ptr Fairness_k_l = bmc_mk_fairness(rbc_fsm, k, l);

  return Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(),  
		    Path_k,
		    Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(),  
			       Loopback_k_l, 
			       Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), 
					  Fairness_k_l, 
					  Tableau_k_l
					 )
			      )
		   );
}


/**Function********************************************************************

  Synopsis           [Returns all problems for the model at length k with 
  all loops from 0 to k]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_MkProblem_k(Fsm_RbcPtr rbc_fsm, node_ptr ltl_wff, int k)
{
  rbc_ptr Path_k;
  rbc_ptr Tableau_k_noloop;
  rbc_ptr Tableau_k_loops = NULL;

  Path_k = Bmc_Mk_k_Path(rbc_fsm, k);

  Tableau_k_noloop = Bmc_MkTableau_k_Noloop(rbc_fsm, ltl_wff, k);

  if (opt_bmc_optimized_tableau(options)) {
    /* use depth1 optimization: */
    node_ptr Fairness = Compile_FsmRbcGetFairness(rbc_fsm);
    int depth =  Bmc_GetLtlWffDepth(ltl_wff);

    if ( (Fairness == NULL) && (depth == 0) ) {
      Tableau_k_loops = Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());
    }
    else if ( (Fairness == NULL) && (depth == 1) ) {
      Tableau_k_loops = bmc_mk_tableau_k_loops_depth1(rbc_fsm, ltl_wff, k);
    }
    else {
      Tableau_k_loops = bmc_mk_tableau_k_loops(rbc_fsm, ltl_wff, k, 0);
    }
    nusmv_assert(Tableau_k_loops != NULL);
  }
  else {
    /* do not use depth1 optimization: */
    Tableau_k_loops = bmc_mk_tableau_k_loops(rbc_fsm, ltl_wff, k, 0);
  }

  return Bmc_RbcAnd( Bmc_VarsMgrGetRbcMgr(), 
		     Path_k,
		     Bmc_RbcOr(Bmc_VarsMgrGetRbcMgr(), 
			       Tableau_k_noloop,
			       Tableau_k_loops) );
}


/**Function********************************************************************

  Synopsis           [Builds and returns the invariant problem of the given propositional formula]

  Description        [Builds the negation of 
                     (I0 imp P0) and ((P0 and R01) imp P1)
		     that must be unsatisfiable.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/

rbc_ptr
Bmc_MkInvariantProblem (Fsm_RbcPtr rbc_fsm, node_ptr wff)
{
  rbc_ptr Init_0 = Bmc_CurrNextShiftAtTime(Compile_FsmRbcGetInit(rbc_fsm), 0);
  rbc_ptr Invar_0 = Bmc_CurrNextShiftAtTime(
			   Compile_FsmRbcGetInvar(rbc_fsm), 0);
  rbc_ptr Trans_0_1_Invar_1 = bmc_mk_j_k_unrolling(rbc_fsm, 0, 1);
  rbc_ptr P = Bmc_Bexp2Rbc(wff);

  rbc_ptr P_0 = Bmc_CurrNextShiftAtTime(P, 0);
  rbc_ptr P_1 = Bmc_CurrNextShiftAtTime(P, 1);

  rbc_ptr Base = Bmc_RbcImplies( Bmc_VarsMgrGetRbcMgr(),
				 Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), 
					    Init_0, 
					    Invar_0),
				 P_0 );

  rbc_ptr Induct = Bmc_RbcImplies( Bmc_VarsMgrGetRbcMgr(),  
				   Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(),  
					    P_0, 
					    Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), 
						       Invar_0, 
						       Trans_0_1_Invar_1)), 
				   P_1 );
  
  return  Bmc_RbcNot( Bmc_VarsMgrGetRbcMgr(), 
		   Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), Base, Induct) );
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/



/**Function********************************************************************

  Synopsis           [Unrolls the transition relation from j to k, taking 
  into account invars over next variables]

  Description        [Using of invars over next variables instead of the 
  previuos variables is a specific implementation aspect]

  SideEffects        []

  SeeAlso            [bmc_mk_and_cnexp_interval]

******************************************************************************/

static rbc_ptr bmc_mk_j_k_unrolling(Fsm_RbcPtr rbc_fsm, int j, int k)
{
  rbc_ptr Trans = Compile_FsmRbcGetTrans(rbc_fsm);
  rbc_ptr Invar = Compile_FsmRbcGetInvar(rbc_fsm);
  rbc_ptr Invar_next = Bmc_CurrShiftToNext(Invar);
  rbc_ptr Trans_Invar_next = Bmc_RbcAnd( Bmc_VarsMgrGetRbcMgr(), 
					 Trans, 
					 Invar_next );
  rbc_ptr Unrolling;

  Unrolling = Bmc_MkAndCurrNextInterval(Trans_Invar_next, j, k - 1);
  return(Unrolling);
}




/**Function********************************************************************

  Synopsis           [Shifts current and next vars in trans expression 
  respectively to k and l steps, if a loop exists. 
  Returns truth if no loop exists]

  Description        []

  SideEffects        []

  SeeAlso            [Bmc_CurrNextShiftAtTimes]

******************************************************************************/

static rbc_ptr bmc_mk_loopback(Fsm_RbcPtr rbc_fsm, int k, int l)
{
  /* checks out the validity of [l, k] only if a loop exists: */
  nusmv_assert((l == BMC_NO_LOOP) || (l <= k));

  if ((l == BMC_NO_LOOP)) {
    return(Bmc_RbcTruth(Bmc_VarsMgrGetRbcMgr()));
  }
  else {
    rbc_ptr Trans = Compile_FsmRbcGetTrans(rbc_fsm);
    if ( Bmc_RbcIsTrue(Bmc_VarsMgrGetRbcMgr(), Trans) ||
	 Bmc_RbcIsFalse(Bmc_VarsMgrGetRbcMgr(), Trans) ) {
      return Trans;
    }

    return ( Bmc_CurrNextShiftAtTimes(Trans, k, l) );
  }
}




/**Function********************************************************************

  Synopsis           [Generates and returns an expression representing 
  all fairnesses in a conjunctioned form]

  Description        [Uses bmc_mk_fairness_aux which recursively call itself
  to conjuctive all fairnesses by constructing a top-level 'and' operation. 
  Moreover bmc_mk_fairness_aux calls the recursive function 
  bmc_mk_one_fairness, which resolves a single fairness as a disjunctioned 
  expression in which each ORed element is a shifting of the single fairness 
  across \[l, k\] if a loop exists. If no loop exists, nothing can be issued, 
  so a falsity value is returned]
 
  SideEffects        []

  SeeAlso            [bmc_mk_fairness_aux, bmc_mk_one_fairness]

******************************************************************************/

static rbc_ptr bmc_mk_fairness(Fsm_RbcPtr rbc_fsm, int k, int l)
{
  node_ptr Fairness = Compile_FsmRbcGetFairness(rbc_fsm);

  return bmc_mk_fairness_aux(Fairness, k, l);
}

static rbc_ptr bmc_mk_fairness_aux(node_ptr fairness, int k, int l)
{
  if (fairness == LS_NIL) {
    return(Bmc_RbcTruth(Bmc_VarsMgrGetRbcMgr()));
  }
  else {
    if (l == BMC_NO_LOOP) {
      return(Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr()));
    }
    else {
      rbc_ptr singleFairness;
      singleFairness = bmc_mk_one_fairness((rbc_ptr)car(fairness), k, l);
      return ( Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), singleFairness,
		      bmc_mk_fairness_aux(cdr(fairness), k, l)) );
    }
  }
}

static rbc_ptr bmc_mk_one_fairness(rbc_ptr one_fairness, int k, int l)
{
  nusmv_assert(l <= k);

  if (l == k) {
    /* exits from the recursive call */ 
    return(Bmc_CurrNextShiftAtTime(one_fairness, l));
  }
  else {
    return ( Bmc_RbcOr(Bmc_VarsMgrGetRbcMgr(), 
		    Bmc_CurrNextShiftAtTime(one_fairness, l),
		    bmc_mk_one_fairness(one_fairness, k, (l + 1))) );
  }
}





/**Function********************************************************************

  Synopsis           [Given time<=k and a \[l, k\] interval, returns next 
  time, or BMC_NO_LOOP if time is equal to k and there is no loop]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
static int bmc_succ_k_l_loop(int time, int k, int l)
{
  nusmv_assert(time <= k);
  
  if (time < k) {
    return(time + 1);
  }
  else {
    /* if there is no loop, l is equal to BMC_NO_LOOP, 
       otherwise time must be resumed to l */
    return(l);
  }
}




/**Function********************************************************************

  Synopsis           [Builds tableau for all possible loops in \[l, k\], 
  taking into account of fairness]

  Description        [Each tableau takes into account of fairnesses relative 
  to its step. All tableau are collected together into a disjunctive form.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static rbc_ptr bmc_mk_tableau_k_loops(Fsm_RbcPtr rbc_fsm, node_ptr ltl_wff, 
				      int k, int l)
{
  rbc_ptr Tableau_k_l;
  rbc_ptr Fairness_k_l;
  rbc_ptr Loopback_k_l;
  rbc_ptr Tableau_LP_FC_k_l;

  /* asserts that given loop l is contained or equal to k step */
  nusmv_assert (!(l == BMC_NO_LOOP));
  nusmv_assert (l <= k);

  Fairness_k_l = bmc_mk_fairness(rbc_fsm, k, l);
  if (Bmc_RbcIsFalse(Bmc_VarsMgrGetRbcMgr(), Fairness_k_l)) {
    return Fairness_k_l;
  }

  Tableau_k_l = Bmc_MkTableau_k_l(ltl_wff, k, l);
  if (Bmc_RbcIsFalse(Bmc_VarsMgrGetRbcMgr(), Tableau_k_l)) {
    return Tableau_k_l;
  }

  Loopback_k_l = bmc_mk_loopback(rbc_fsm, k, l);
  if (Bmc_RbcIsFalse(Bmc_VarsMgrGetRbcMgr(), Loopback_k_l)) {
    return Loopback_k_l;
  }

  /* tableau + loopback + fairness at step k with loop at l = k */
  Tableau_LP_FC_k_l = Bmc_RbcAnd( Bmc_VarsMgrGetRbcMgr(),  
				  Tableau_k_l, 
				  Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), 
					     Fairness_k_l, 
					     Loopback_k_l) 
				); 

  if (k == l) { 
    /* tableau + loopback at step k with loop at l = k */
    return Tableau_LP_FC_k_l;
  }
  else { 
    /* OR of tableau + loopback + fairness at step k with loop at l,
       and for the (tableau + loopback) at k step and loop at l + 1 ... k */
     return Bmc_RbcOr(Bmc_VarsMgrGetRbcMgr(), 
		      Tableau_LP_FC_k_l, 
		      bmc_mk_tableau_k_loops(rbc_fsm, ltl_wff, k, l + 1)
		     );
  }
}


/**Function********************************************************************

  Synopsis           [Builds tableau for all possible loops in \[l, k\], in 
  the particular case in which depth is 1. This function takes into account 
  of fairness]

  Description        [Builds the tableau in the case depth==1 as suggested 
  by Sebastiani]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static rbc_ptr 
bmc_mk_tableau_k_loops_depth1(Fsm_RbcPtr rbc_fsm, node_ptr ltl_wff, int k)
{
  rbc_ptr result=NULL;

  switch (node_get_type (ltl_wff)) {
  case OP_FUTURE:                       /* OP_FUTURE */
  case UNTIL:                           /* UNTIL     */
    result = Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());
    break;
    
  case OP_NEXT:                         /* OP_NEXT   */
    if (k==0) {
      result = Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(),
			  Bmc_MkTableau_k_l(car(ltl_wff), k, 0),
			  bmc_mk_loopback(rbc_fsm, k, 0));
    }
    else {
      result = Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());
    }
    break;
  
  case OP_GLOBAL:                       /* OP_GLOBAL */
    result = Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(),
			bmc_mk_all_loops(rbc_fsm, k),
			Bmc_MkGloballyTableauAtTime(car(ltl_wff), 0, k, 0));
    break;

  case RELEASES:                        /* RELEASES     */
    result = Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(),
			bmc_mk_all_loops(rbc_fsm, k),
			Bmc_MkGloballyTableauAtTime(cdr(ltl_wff), 0, k, 0));
    break;

  case AND:                             /* AND       */
  case OR:                              /* OR        */
  case IFF:                             /* IFF        */
    result = Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(),
			bmc_mk_all_loops(rbc_fsm, k),
			Bmc_MkTableauAtTime(ltl_wff, 0, k, 0));
    break;

  case IMPLIES:
    internal_error("Unexpected 'Implies': it should have been nnf-ed away!\n");

  case TRUEEXP:                         /* TRUEEXP   */
  case FALSEEXP:                        /* FALSEEXP  */
  case DOT:                             /* VARIABLE  */
  case NOT:                             /* LITERAL      */
    internal_error("Unexpected formula with depth zero had been found.\n");

  case ATOM:
  case ARRAY:
  case EX:
  case AX:
  case EF:
  case AF:
  case EG:
  case AG:
  case EE:
  case AA:
  case EBF:
  case EBG:
  case ABF:
  case ABG:
  case BUNTIL:
  case MMIN:
  case MMAX:
  case APATH:
  case EPATH:
    internal_error( "Unexpected CTL operator, node type %d\n",
		    node_get_type(ltl_wff) );

  default:
    /* no other types is expected here */
    nusmv_assert( false );
  }

  nusmv_assert(result != NULL); /*it must be assigned! */
  return result;
}

/**Function********************************************************************

  Synopsis           [auxiliary part of Bmc_MkUntilTableauAtTime]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
static rbc_ptr
bmc_mk_untiltableau_at_time_aux(node_ptr p, node_ptr q, int time,
				int k, int l, int steps)
{
  rbc_ptr tableau_temp; /* for increasing of performances */ 
  rbc_ptr tableau_following; /* to increase readability */

  nusmv_assert (time <= k);
  nusmv_assert (steps >= 1);
  
  tableau_temp = Bmc_MkTableauAtTime(q, time, k, l); 

  if (steps > 1) {
    tableau_following = bmc_mk_untiltableau_at_time_aux( p, q, 
			    bmc_succ_k_l_loop(time, k, l),
			    k, l, steps - 1 );

    tableau_temp = Bmc_RbcOr( Bmc_VarsMgrGetRbcMgr(),  
			      tableau_temp,
			      Bmc_RbcAnd
			      ( Bmc_VarsMgrGetRbcMgr(), 
				Bmc_MkTableauAtTime(p, time, k, l), 
				tableau_following 
			      )  
			    );
  }

  return tableau_temp;

}




/**Function********************************************************************

  Synopsis           [auxiliary part of Bmc_MkReleasesTableauAtTime]

  Description        [Builds the release operator expression]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static rbc_ptr
bmc_mk_releasestableau_at_time_aux(node_ptr p, node_ptr q, int time,
				   int k, int l, int steps)
{
  rbc_ptr tableau_p;
  rbc_ptr tableau_q;
  rbc_ptr tableau_result;
  
  nusmv_assert (time <= k);
  nusmv_assert (steps >= 1);

  tableau_p = Bmc_MkTableauAtTime(p, time, k, l);
  tableau_q = Bmc_MkTableauAtTime(q, time, k, l);

  if (steps == 1) {
    if (l == BMC_NO_LOOP) { /* q & p */
      tableau_result = Bmc_RbcAnd(Bmc_VarsMgrGetRbcMgr(), 
                                  tableau_p, 
                                  tableau_q);
    } else { /* q */
      tableau_result = tableau_q;
    }
  }
  else { /* q & ( p | X(p V q) ) */
    rbc_ptr 
    tableau_following = 
      bmc_mk_releasestableau_at_time_aux( p, q,
                                          bmc_succ_k_l_loop(time, k, l),
                                          k, l, steps - 1
                                          );
    tableau_result = Bmc_RbcAnd( Bmc_VarsMgrGetRbcMgr(),  
                                 tableau_q, 
                                 Bmc_RbcOr(Bmc_VarsMgrGetRbcMgr(), 
                                           tableau_p, 
                                           tableau_following)
                                 );
  }

  return tableau_result;
}


/**Function********************************************************************

  Synopsis           [Builds the disjunction of all the loops from 0 to k]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static rbc_ptr bmc_mk_all_loops(Fsm_RbcPtr rbc_fsm, int k)
{
  int l;
  rbc_ptr result=Bmc_RbcFalsity(Bmc_VarsMgrGetRbcMgr());

  for (l=0; l<=k; ++l) {
    result = Bmc_RbcOr( Bmc_VarsMgrGetRbcMgr(),
                        bmc_mk_loopback(rbc_fsm, k, l),
                        result );
  }

  return result;
}

