/**CFile***********************************************************************

  FileName    [bmcUtils.c]

  PackageName [bmc]

  Synopsis    [Useful internal functions of the bmc package]

  Description []

  SeeAlso     []

  Author      [Roberto Cavada]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/



/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static void freeListOfLists_aux ARGS((lsList list));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Initializes the BMC model manager structure]

  Description        []

  SideEffects        [...]

  SeeAlso            []
******************************************************************************/
void Bmc_Init()
{
  /* Hash table for shifting operation: */
  BmcShiftingHashInit();


  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "Initializing BMC variables manager ... \n");
  }
  /* Vars manager: */  
  BmcVarsMgrInit(get_boolean_variable_list());

}

/**Function********************************************************************

  Synopsis           [Frees all resources allocated for the BMC model manager]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
void Bmc_Quit(void)
{
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "\nResetting BMC module ... \n");
  }

  BmcVarsMgrQuit();

  /* Resets the _bmc_test_tableau command private data: */ 
  Bmc_TestReset(); 

  /* Hash table for shifting operation: */
  BmcShiftingHashQuit();
}



/**Function********************************************************************

  Synopsis           [Given a string representing a loopback possible value, 
  returns the integer corresponding value as parameter. Returns SUCCESS if 
  conversion was possible, SYNTAX_ERROR otherwise]

  Description        [Use this function to correctly convert a string 
  containing a loopback user-side value to the internal representation of
  the same loopback value] 

  SideEffects        [Integer value passed as argumnet will change]

  SeeAlso            []

******************************************************************************/
outcome Bmc_ConvertLoopbackFromString(const char* strValue, int* iValue) 
{
  int k = get_bmc_pb_length(options);
  int l = 0;
  outcome ret = SUCCESS;

  if (strValue == NIL(char)) return SYNTAX_ERROR;

  if (BMC_IS_ALL_LOOPBACKS(strValue))     l = BMC_ALL_LOOPS;
  else if (BMC_IS_NO_LOOPBACK(strValue))  l = BMC_NO_LOOP;
  else if (util_str2int(strValue, &l) == 0) {
    if ( (l == BMC_ALL_LOOPS) || (l == BMC_NO_LOOP)) {
      /* user supplied a private integer value: */
      ret = SYNTAX_ERROR; 
    } 
  }
  else ret = SYNTAX_ERROR;
  
  /* checks loopback value, and returns it if it is ok */
  if (ret == SUCCESS) {
    /* l must be in k */
    if ( (l==BMC_ALL_LOOPS) || (l==BMC_NO_LOOP) || (abs(l)<=k) ) {
      *iValue = l;
    }
    else ret = SYNTAX_ERROR;
  }

  return ret;
}  


/**Function********************************************************************

  Synopsis           [Given an integer containing the inner representation 
  of the loopback value, returns as parameter the corresponding user-side 
  value as string.  
  ]

  Description        [Inverse semantic of Bmc_ConvertLoopbackFromString. 
  bufsize is the maximum buffer size]

  SideEffects        [String buffer passed as argument will change]

  SeeAlso            [Bmc_ConvertLoopbackFromString]

******************************************************************************/
void Bmc_ConvertLoopbackFromInteger(const int iLoopback, char* szLoopback, 
				    const int _bufsize) 
{
  int iCheck; /* for buffer operations checking only */
  int bufsize = _bufsize-2; /* to store terminator */

  nusmv_assert(bufsize > 0); /* one character+terminator at least! */
  szLoopback[bufsize+1] = '\0'; /* put terminator */
  
  if (iLoopback == BMC_ALL_LOOPS) {
    iCheck = snprintf(szLoopback, bufsize, "%s", 
		      BMC_ALL_LOOPS_USERSIDE_SYMBOL);
    nusmv_assert(iCheck>=0);
  }

  else if (iLoopback == BMC_NO_LOOP) {
    iCheck = snprintf(szLoopback, bufsize, "%s", BMC_NO_LOOP_USERSIDE_SYMBOL);
    nusmv_assert(iCheck>=0);
  }
  else {
    /* value is ok, convert to string: */
    iCheck = snprintf(szLoopback, bufsize, "%d", iLoopback);
    nusmv_assert(iCheck>=0);
  }

}



/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Return pathname without path prefix]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
char* Bmc_StripPath(char * pathfname)
{
  char* old_pos = pathfname;

  while ((*pathfname) != '\0') {
    if ((*pathfname) == '/') old_pos = pathfname + 1;
    
    ++pathfname; 
  }

  return old_pos;
}


/**Function********************************************************************

  Synopsis           [Returns filename without path and extension]

  Description        [Example: given "~/.../test.smv", "test" will be returned.
  filename must be a string whose length is large enought to contain the "pure"
  filename]

  SideEffects        [the string pointed by 'filename' changes]

  SeeAlso            []

******************************************************************************/
void Bmc_StripPathExt(char * fpathname, char* filename)
{
  char* szNoPath = NULL;
  char* szExt = NULL;

  szNoPath = Bmc_StripPath(fpathname);
  szExt = strstr(szNoPath, ".");
  if (szExt != NULL) {
    strncpy(filename, szNoPath, (size_t)(szExt - szNoPath));
    *(filename+(size_t)(szExt - szNoPath)) = '\0'; /* terminates the string */
  }
  else {
    strcpy(filename, szNoPath);
  }
}




/**Function********************************************************************

  Synopsis           [Destroys a list of list]

  Description        [This function can be used to destroy lists of list. The 
  contained set of lists is removed from memory as the top level list. 
  More than two levels are not handled at the moment.]

  SideEffects        [Lists are deallocated]

  SeeAlso            [lsDestroy]

******************************************************************************/
void BmcIntFreeListOfLists(lsList list_of_lists) 
{
  lsDestroy(list_of_lists, &freeListOfLists_aux);
}
  


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Private service for BmcIntFreeListOfLists]

  SideEffects        []

  SeeAlso            [BmcIntFreeListOfLists]

******************************************************************************/
static void freeListOfLists_aux(lsList list)
{
  lsDestroy(list, NULL);
}

