/**CFile***********************************************************************

  FileName    [bmcWff.c]

  PackageName [bmc]

  Synopsis    [Well Formed Formula manipulation routines]

  Description []

  SeeAlso     []

  Author      [Alessandro Cimatti and Lorenzo Delana]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static node_ptr ltl_wff_expand_case_aux ARGS((node_ptr wff));
static node_ptr ltl_wff_expand_case     ARGS((node_ptr wff));


/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Make a <b>binary</b> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkBinary(int type, node_ptr arg1, node_ptr arg2)
{
  return find_node(type, arg1, arg2);
}

/**Function********************************************************************

  Synopsis           [Make a <b>unary</b> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkUnary(int type, node_ptr arg)
{
  return find_node (type, arg, Nil);
}

/**Function********************************************************************

  Synopsis           [Make a <b>constant</b> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkConst(int type)
{
  return find_node (type, Nil, Nil);
}

/**Function********************************************************************

  Synopsis           [Make a <i>truth</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkTruth(void)
{
  return Bmc_WffMkConst(TRUEEXP);
}

/**Function********************************************************************

  Synopsis           [Make a <i>false</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkFalsity(void)
{
  return Bmc_WffMkConst(FALSEEXP);
}

/**Function********************************************************************

  Synopsis           [Make a <i>not</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_WffMkNot(node_ptr arg)
{
  return Bmc_WffMkUnary(NOT, arg);
}

/**Function********************************************************************

  Synopsis           [Make an <i>and</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkAnd(node_ptr arg1, node_ptr arg2)
{
  return Bmc_WffMkBinary(AND, arg1, arg2);
}

/**Function********************************************************************

  Synopsis           [Make an <i>or</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkOr(node_ptr arg1, node_ptr arg2)
{
  return Bmc_WffMkBinary(OR, arg1, arg2);
}

/**Function********************************************************************

  Synopsis           [Make an <i>implies</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkImplies(node_ptr arg1, node_ptr arg2)
{
  return Bmc_WffMkBinary(IMPLIES, arg1, arg2);
}

/**Function********************************************************************

  Synopsis           [Make an <i>iff</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkIff(node_ptr arg1, node_ptr arg2)
{
  return Bmc_WffMkBinary(IFF, arg1, arg2);
}

/**Function********************************************************************

  Synopsis           [Make a <i>next</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_WffMkNext(node_ptr arg)
{
  return Bmc_WffMkUnary(NEXT, arg);
}

/**Function********************************************************************

  Synopsis           [Applies <i>op_next</i> x times]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_WffMkXopNext(node_ptr arg, int x)
{
  nusmv_assert (x >= 0);

  if (x == 0)
    return arg;
  else
    return Bmc_WffMkOpNext(Bmc_WffMkXopNext(arg, x - 1));
}

/**Function********************************************************************

  Synopsis           [Make an <i>op_next</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_WffMkOpNext(node_ptr arg)
{
  return Bmc_WffMkUnary(OP_NEXT, arg);
}

/**Function********************************************************************

  Synopsis           [Make a <i>globally</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkGlobally(node_ptr arg)
{
  return Bmc_WffMkUnary(OP_GLOBAL, arg);
}

/**Function********************************************************************

  Synopsis           [Make an <i>eventually</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkEventually(node_ptr arg)
{
  return Bmc_WffMkUnary(OP_FUTURE, arg);
}

/**Function********************************************************************

  Synopsis           [Make an <i>until</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkUntil(node_ptr arg1, node_ptr arg2)
{
  return Bmc_WffMkBinary(UNTIL, arg1, arg2);
}

/**Function********************************************************************

  Synopsis           [Make a <i>releases</i> WFF]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
node_ptr
Bmc_WffMkReleases(node_ptr arg1, node_ptr arg2)
{
  return Bmc_WffMkBinary(RELEASES, arg1, arg2);
}


/**Function********************************************************************

  Synopsis           [Make the <b>negative normal form</b> of given WFF]

  Description        [a positive (1) polarity will not negate entire formula]

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/

node_ptr
Bmc_LtlWffMkNnfToplevel(node_ptr wff) {
  return Bmc_LtlWffMkNnf(wff, true);
}


node_ptr
Bmc_LtlWffMkNnf (node_ptr wff, boolean pol)
{
  /* if reached a Nil branch then end recursion with a Nil node */
  if (wff == Nil) return Nil;

  /* Only temporal operator X (node type OP_NEXT) is legal in LTL wffs.
     The operator NEXT used in model definition is trapped. */
  nusmv_assert (node_get_type (wff) != NEXT);

  switch (node_get_type (wff)) {
  case TRUEEXP:                         /* TRUEEXP   */
    if (pol == true)
      return Bmc_WffMkTruth();
    else
      /* !1 <-> 0 */
      return Bmc_WffMkFalsity();
  case FALSEEXP:                        /* FALSEEXP  */
    if (pol == true)
      return Bmc_WffMkFalsity();
    else
      /* !0 <-> 1 */
      return Bmc_WffMkTruth();
  case NOT:                             /* NOT       */
    /* !(a) <-> (!a) */
    /* !(!a) <-> a */
    return Bmc_LtlWffMkNnf(car (wff), !pol);
  case AND:                             /* AND       */
    if (pol == true)
      return Bmc_WffMkAnd(Bmc_LtlWffMkNnf(car (wff), true),
			  Bmc_LtlWffMkNnf(cdr (wff), true));
    else
      /* !(a & b) <-> (!a | !b) */
      return Bmc_WffMkOr(Bmc_LtlWffMkNnf(car (wff), false),
			 Bmc_LtlWffMkNnf(cdr (wff), false));
  case OR:                              /* OR        */
    if (pol == true)
      return Bmc_WffMkOr(Bmc_LtlWffMkNnf(car (wff), true),
			 Bmc_LtlWffMkNnf(cdr (wff), true));
    else
      /* !(a | b) <-> (!a & !b) */
      return Bmc_WffMkAnd(Bmc_LtlWffMkNnf(car (wff), false),
			  Bmc_LtlWffMkNnf(cdr (wff), false));
  case IMPLIES:                         /* IMPLIES   */
    if (pol == true)
      /* (a -> b) <-> !(a & !b) <-> (!a | b) */
      return Bmc_WffMkOr(Bmc_LtlWffMkNnf(car (wff), false),
			 Bmc_LtlWffMkNnf(cdr (wff), true));
    else
      /* !(a -> b) <-> (a & !b) */
      return Bmc_WffMkAnd(Bmc_LtlWffMkNnf(car (wff), true),
			  Bmc_LtlWffMkNnf(cdr (wff), false));

  case IFF:                             /* IFF       */
    if (pol == true)
      /* (a <-> b) <->
	 !(a & !b) & !(b & !a) <->
	 (!a | b) & (!b | a) */
      return Bmc_WffMkAnd( Bmc_WffMkOr(Bmc_LtlWffMkNnf(car (wff), false),
				       Bmc_LtlWffMkNnf(cdr (wff), true)),
			   Bmc_WffMkOr(Bmc_LtlWffMkNnf(car (wff), true),
				       Bmc_LtlWffMkNnf(cdr (wff), false))
			 );
    else
      /* !(a <-> b) <->
	 !(!(a & !b) & !(b & !a)) <->
	 (a & !b) | (b & !a) */
      return Bmc_WffMkOr(Bmc_WffMkAnd(Bmc_LtlWffMkNnf(car (wff), true),
				      Bmc_LtlWffMkNnf(cdr (wff), false)),
			 Bmc_WffMkAnd(Bmc_LtlWffMkNnf(car (wff), false),
				      Bmc_LtlWffMkNnf(cdr (wff), true))
		       );

  case OP_NEXT:                         /* OP_NEXT   */
    /* !X(a) <-> X(!a) */
    return Bmc_WffMkOpNext(Bmc_LtlWffMkNnf(car (wff), pol));

  case OP_GLOBAL:                       /* OP_GLOBAL */
    if (pol == true)
      return Bmc_WffMkGlobally(Bmc_LtlWffMkNnf(car (wff), pol));
    else
      /* !G(a) <-> F(!a) */
      return Bmc_WffMkEventually(Bmc_LtlWffMkNnf(car (wff), pol));

  case OP_FUTURE:                       /* OP_FUTURE */
    if (pol == true)
      return Bmc_WffMkEventually(Bmc_LtlWffMkNnf(car (wff), pol));
    else
      /* !F(a) <-> G(!a) */
      return Bmc_WffMkGlobally(Bmc_LtlWffMkNnf(car (wff), pol));

  case UNTIL:                           /* UNTIL     */
    if (pol == true)
      return Bmc_WffMkUntil(Bmc_LtlWffMkNnf(car (wff), pol),
			    Bmc_LtlWffMkNnf(cdr (wff), pol));
    else
      /* !(a U b) <-> (!a V !b) */
      return Bmc_WffMkReleases(Bmc_LtlWffMkNnf(car (wff), pol),
			       Bmc_LtlWffMkNnf(cdr (wff), pol));

  case RELEASES:                        /* RELEASES  */
    if (pol == true)
      return Bmc_WffMkReleases(Bmc_LtlWffMkNnf(car (wff), pol),
			       Bmc_LtlWffMkNnf(cdr (wff), pol));
    else
      /* !(a V b) <-> (!a U !b) */
      return Bmc_WffMkUntil(Bmc_LtlWffMkNnf(car (wff), pol),
			    Bmc_LtlWffMkNnf(cdr (wff), pol));

  case CASE: 
    {
      node_ptr nocase_wff = ltl_wff_expand_case(wff);
      return Bmc_LtlWffMkNnf(nocase_wff, pol);
    }

  case DOT: /* is a bexp var */
    if (pol == true)
      return wff;
    else
      return Bmc_WffMkNot(wff);

  case ARRAY:  /* internal format for atoms that */
  case NUMBER: /* should have been previously hidden within DOT */
  case ATOM: 
    internal_error("Bmc_LtlWffMkNnf: unexpected node %s\n",
	   (node_get_type(wff) == ARRAY ? "ARRAY" :
	    (node_get_type(wff) == NUMBER ? "NUMBER" : "ATOM")));
    return Nil;

  default: /* (default action) */
    internal_error("Bmc_LtlWffMkNnf: unexpected TOKEN %d\n", 
		   node_get_type(wff));
    return Nil;
  }
}




/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis           [Returns the modal depth of the given formula]

  Description        [Returns 0 for propositional formulae, 1 or more for 
  temporal formulae]

  SideEffects        [none]

  SeeAlso            []

******************************************************************************/
int Bmc_GetLtlWffDepth(node_ptr ltl_wff)
{
  int depth = -1;
  int d1,d2;

  switch (node_get_type (ltl_wff)) {
  case TRUEEXP:                           /* TRUEEXP   */
  case FALSEEXP:                          /* FALSEEXP  */
    depth = 0;
    break;
    
  case NOT:                             /* NOT       */
    depth = Bmc_GetLtlWffDepth(car(ltl_wff));
    break;
    
  case AND:                             /* AND       */
  case OR:                              /* OR        */
  case IFF:                             /* IFF        */
    d1 = Bmc_GetLtlWffDepth(car(ltl_wff));
    d2 = Bmc_GetLtlWffDepth(cdr(ltl_wff));
    depth = max(d1,d2);
    break;
    
  case OP_NEXT:                         /* OP_NEXT   */
  case OP_GLOBAL:                       /* OP_GLOBAL */
  case OP_FUTURE:                       /* OP_FUTURE */
    depth = 1 + Bmc_GetLtlWffDepth(car(ltl_wff));
    break;
    
  case UNTIL:                           /* UNTIL     */
  case RELEASES:                        /* RELEASES  */
    d1 = Bmc_GetLtlWffDepth(car(ltl_wff));
    d2 = Bmc_GetLtlWffDepth(cdr(ltl_wff));
    depth = 1 + max (d1,d2);
    break;
    
  case IMPLIES:
    internal_error("implies should have been nnf-ef away!\n");
    break;
    
  case DOT:
    depth = 0;
    break;

  case ATOM:
  case ARRAY:
  case EX:
  case AX:
  case EF:
  case AF:
  case EG:
  case AG:
  case EE:
  case AA:
  case EBF:
  case EBG:
  case ABF:
  case ABG:
  case BUNTIL:
  case MMIN:
  case MMAX:
  case APATH:
  case EPATH:
    internal_error( "Unexpected CTL operator, node type %d\n",
		    node_get_type(ltl_wff) );
 
  default:                              
    /* no other cases are currently allowed */
    nusmv_assert(FALSE);
  }

  return depth;
}


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

static node_ptr ltl_wff_expand_case(node_ptr wff) {
  node_ptr res;

  nusmv_assert (node_get_type (wff) == CASE);
  res = ltl_wff_expand_case_aux(wff);

  return(res);
}

static node_ptr ltl_wff_expand_case_aux(node_ptr wff) {
  if (node_get_type(wff) == CASE) {
    node_ptr cur_cond  = car(car(wff));
    node_ptr cur_rslt  = cdr(car(wff));
    node_ptr case_rest = cdr(wff);
    node_ptr res;

    nusmv_assert (node_get_type(car(wff)) == COLON);

    res = Bmc_WffMkOr( Bmc_WffMkAnd(cur_cond, cur_rslt),
		       Bmc_WffMkAnd(Bmc_WffMkNot(cur_cond),
				    ltl_wff_expand_case_aux(case_rest))
		     );
    return(res);

  }
  else {
    nusmv_assert((node_get_type(wff) == TRUEEXP) || 
		 (node_get_type(wff) == FALSEEXP));

    return(wff);
  }
}


