/**CFile***********************************************************************

  FileName    [compileFsmUtil.c]

  PackageName [compile]

  Synopsis    [general functions for all conjunctive partitioning methods]

  Description [general functions for all conjunctive partitioning methods]

  SeeAlso     []

  Author      [Emanuele Olivetti]

  Copyright   [
  This file is part of the ``compile'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "compileInt.h" 

static char rcsid[] UTIL_UNUSED = "$Id: compileFsmUtil.c,v 1.18 2001/11/05 10:57:02 none Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
/**Variable********************************************************************

  Synopsis    [Hash to keep track of already considered expressions.]

  Description [Hash to keep track of already considered expressions.]

  SeeAlso     []

******************************************************************************/
static st_table * CP_hash = (st_table*)NULL;

static void init_CP_hash() 
{
  CP_hash = st_init_table(st_ptrcmp, st_ptrhash);
  if (CP_hash == (st_table *)NULL) {
    fprintf(nusmv_stderr, "init_CP_hash: unable to allocate local hash.\n");
    nusmv_exit(1);
  }
}
static void quit_CP_hash()
{
  nusmv_assert(CP_hash != (st_table *)NULL);
  st_free_table(CP_hash);
}
static void insert_CP_hash(node_ptr el)
{
  if (st_add_direct(CP_hash, (char *)el, (char *)1) == ST_OUT_OF_MEM) {
    fprintf(nusmv_stderr, "insert_CP_hash: Unable to insert result in local hash.\n");
    nusmv_exit(1);
  }
}
static int query_CP_hash(node_ptr el)
{
  int v;
  int result = st_lookup(CP_hash, (char *)el, (char **)&v);

  return(result);
}


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static node_ptr compileBuildBasicCPRecur ARGS((node_ptr ,node_ptr ,add_ptr));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Allocates and initializes memory for CPCluster_Info struct.]

  Description        [Allocates and initializes memory for CPCluster_Info struct.]

  SideEffects        []

******************************************************************************/

CPCluster_Info_t * CPClusterInfoAlloc(void)
{
  CPCluster_Info_t * CI;

  CI = ALLOC(CPCluster_Info_t, 1);
  CI->Ti          = (bdd_ptr)NULL;
  CI->Ei          = (bdd_ptr)NULL;
  return(CI);
}

/**Function********************************************************************

  Synopsis           [Frees the  memory allocated for CPCluster_Info struct.]

  Description        [Frees the  memory allocated for CPCluster_Info struct.]

  SideEffects        []

******************************************************************************/
void CPClusterInfoFree(DdManager *dd, CPCluster_Info_t * CI)
{
  if (CI != (CPCluster_Info_t *)NULL) {
    if (CI->Ti        != (bdd_ptr)NULL) bdd_free(dd, CI->Ti);
    if (CI->Ei        != (bdd_ptr)NULL) bdd_free(dd, CI->Ei);
    FREE(CI);
  }
}

/**Function********************************************************************

  Synopsis           [Frees a list of CPCluster_Info struct.]

  Description        [Frees a list of CPCluster_Info struct.]

  SideEffects        []

******************************************************************************/
void CPFreeClustersList(DdManager *dd, node_ptr CL)
{
  node_ptr l = CL;

  while (l != Nil) {
    CPCluster_Info_t * c = (CPCluster_Info_t *)car(l);
    CPClusterInfoFree(dd, c);
    l = cdr(l);
  }
  free_list(CL);
}

/**Function********************************************************************

  Synopsis          [Duplicates a list of cluster allocating necessary memory.]

  Description       [Duplicates a list of cluster allocating necessary memory.]

  SideEffects       []

******************************************************************************/
void CPDupClustersList(DdManager *dd, node_ptr CL, node_ptr *newCL)
{
  node_ptr l = CL;
  *newCL = Nil;

  while (l != Nil) {
    CPCluster_Info_t * c = (CPCluster_Info_t *)car(l);
    CPCluster_Info_t * newc = CPClusterInfoAlloc();
    if (c->Ti !=NULL) {
      newc->Ti = bdd_dup (c->Ti);
    } else {
      newc->Ti = NULL;
    }
    if (c->Ei !=NULL) {
      newc->Ei = bdd_dup (c->Ei);
    } else {
      newc->Ei = NULL;
    }
    *newCL = cons((node_ptr)newc,*newCL);
    l = cdr(l);
  }
  *newCL=reverse(*newCL);
}

/**Function********************************************************************

  Synopsis           [Builds BDD list from given variables list.]

  Description        [Builds BDD list from given variables list.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
node_ptr compileBuildElementaryBDDList(node_ptr list_variables, add_ptr assumption)
{
  node_ptr l;
  node_ptr cluster_bdd = Nil;
  init_CP_hash();
  for(l = list_variables; l != Nil; l = cdr(l)){
    node_ptr var_i = car(l);
    node_ptr next_i = lookup_variable_sexp_model_hash(var_i);

    if (opt_verbose_level_gt(options, 2)) {
      indent_node(nusmv_stderr, " computing next values for variable: ",
                  var_i, "\n");
    }
    cluster_bdd = compileBuildBasicCPRecur(cluster_bdd,
                                               var_model_sexp_get_next(next_i),
                                               assumption);
  }
  quit_CP_hash();
  cluster_bdd = reverse(cluster_bdd);
  return(cluster_bdd);
}


/**Function********************************************************************

  Synopsis           [Checks that the And_i Ti == Mon_T]

  Description        [Checks whether the conjunction of the transition
                      relations of all the clusters equals the the
                      monolithic Transition relation.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static int CPCheckMonolithic(DdManager *dd, node_ptr Clist, bdd_ptr Mon_T)
{
  bdd_ptr Acc = bdd_one(dd);
  int result = 0;

  while(Clist != Nil) {
    CPCluster_Info_t * Ci = (CPCluster_Info_t *)car(Clist);

    bdd_and_accumulate(dd, &Acc, Ci->Ti);
    
    Clist = cdr(Clist);
  }
  result = (Acc == Mon_T);
  bdd_free(dd, Acc);
  return(result);
}

/**Function********************************************************************

  Synopsis           [Perform some test on transition relation]

  Description        [Check the equality with monolithic transition relation,
                      and check the quantification schedule.]

  SideEffects        []

  SeeAlso            [CPCheckMonolithic, CPCheckQuantificationSchedule]

******************************************************************************/
void compileCPCheckPartitionedTransition(CP_Ptr transition, node_ptr list_variables, bdd_ptr assumption)
{
  CP_Ptr mono = Compile_BuildModelMonolithic(list_variables, assumption);
  int i = 0;
  fprintf(nusmv_stderr,
	  "\nChecking equality Partitioned vs. Monolithic : \n");
  i = CPCheckMonolithic(dd_manager, CP_GetBackwardTrans(transition),
			((CPCluster_Info_t *)car(CP_GetBackwardTrans(mono)))->Ti);
  fprintf(nusmv_stderr,"\tBackward : %d\n",i);
  i = CPCheckMonolithic(dd_manager, CP_GetForwardTrans(transition),
			((CPCluster_Info_t *)car(CP_GetForwardTrans(mono)))->Ti);
  fprintf(nusmv_stderr,"\tForward  : %d\n",i);

  fprintf(nusmv_stderr,"\nChecking Quantification Schedule : \n");
  i = CPCheckQuantificationSchedule(dd_manager, CP_GetBackwardTrans(transition));
  fprintf(nusmv_stderr,"\tBackward : %d\n",i);
  i = CPCheckQuantificationSchedule(dd_manager, CP_GetForwardTrans(transition));
  fprintf(nusmv_stderr,"\tForward  : %d\n",i);
}

/**Function********************************************************************

  Synopsis    [Given a list of clusters representing the relations this
               function checks if the schedule is correct.]

  Description [Let Ci and Ti be the ith cube and relation in the list.
               The schedule is correct iff<br>
  <ol>
  <li> For all Tj: j > i, S(Tj) and S(Ci) do not intersect, i.e., the
       variables which are quantified in Ci should not appear in the
       Tj for j>i.</li>
  <li> For any j, S(Ci) and S(Cj) do not intersect.</li>
  </ol><br>

  where S(T) is the set of support of the BDD T.
  Returns 1 if the schedule is correct, 0 otherwise.
  This function is implemented for checking the correctness of the
  clustering algorithm only.]

  SideEffects []

******************************************************************************/
int CPCheckQuantificationSchedule(DdManager *dd, node_ptr Cl_list)
{
  node_ptr Clist = Cl_list;
  
  while (Clist != Nil) {
    CPCluster_Info_t * Cj = (CPCluster_Info_t *)car(Clist);
    node_ptr Nlist = cdr(Clist);

    while (Nlist!= Nil) {
      CPCluster_Info_t * Ci = (CPCluster_Info_t *)car(Nlist);
      /* bdd_ptr intersection = bdd_cube_diff(dd, Cj->Ei, Ci->Supp_Ti); */
      bdd_ptr intersection = bdd_cube_diff(dd, Cj->Ei, bdd_support(dd,Ci->Ti));

      bdd_free(dd, intersection);
      if (intersection !=  Cj->Ei)
        /* Violates the condition "a" */
        return(-1);
      intersection = bdd_cube_diff(dd, Cj->Ei,  Ci->Ei);
      bdd_free(dd, intersection);
      if ((intersection !=  Cj->Ei) && (get_partition_method(options)==Iwls95CP))
        /* Violates the condition "b" */
        return(-2);
      Nlist = cdr(Nlist);
    } /* while (Nlist!= Nil) */
    Clist = cdr(Clist);
  } /* while (Clist != Nil) */
  return(1);
}

/**Function********************************************************************

  Synopsis           [Returns a list with one new cluster and computes its quantifiers.]

  Description        [Returns a list with one new cluster and computes its quantifiers.]

  SideEffects        [required]

  SeeAlso            [optional]

******************************************************************************/
node_ptr AddOneCluster(CPCluster_Info_t * Ci, node_ptr old_clusters, bdd_ptr vars)
{
  node_ptr new_clusters,l;
  bdd_ptr tmp,tmp2;
  CPCluster_Info_t * newCi;
  
  newCi = CPClusterInfoAlloc();
  newCi->Ti = bdd_dup(Ci->Ti);

  CPDupClustersList(dd_manager, old_clusters, &new_clusters);

  tmp = vars;
  l = new_clusters;
  while (l!=Nil) {
    CPCluster_Info_t * Ci = (CPCluster_Info_t *)(car(l));
    tmp2 = bdd_cube_diff(dd_manager, tmp, Ci->Ei);
    bdd_free(dd_manager,tmp);
    tmp = tmp2;
    l = cdr(l);
  }
  newCi->Ei = tmp;
  new_clusters = cons((node_ptr)newCi, new_clusters);

  return(new_clusters);
}


/**Function********************************************************************

  Synopsis           [Add the LTL tableau as a new BDD cluster in the transition relation.]

  Description        [Add the LTL tableau as a new BDD cluster in the transition relation.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
CP_Ptr Compile_CPAddClusterAndFillSchedule(CP_Ptr trans, bdd_ptr tableau_bdd)
{
  node_ptr fwd_clusters , bwd_clusters, old;
  bdd_ptr vars;
  CPCluster_Info_t * Ci;
  CP_Ptr result;
  
  result = CP_PtrAlloc();
  nusmv_assert(result != (CP_Ptr)NULL);

  Ci = CPClusterInfoAlloc();

  /*  Ci->Ti = add_to_bdd(dd_manager, tableau_add);*/
  Ci->Ti = bdd_dup(tableau_bdd);

  vars = bdd_and(dd_manager, state_variables_bdd, input_variables_bdd); 
  old = CP_GetForwardTrans(trans);
  fwd_clusters = AddOneCluster(Ci, old, vars);
  CP_SetForwardTrans(result, fwd_clusters);

  vars = bdd_and(dd_manager, next_state_variables_bdd, next_input_variables_bdd);
  old = CP_GetBackwardTrans(trans);
  bwd_clusters = AddOneCluster(Ci, old, vars);
  CP_SetBackwardTrans(result, bwd_clusters);

  CPClusterInfoFree(dd_manager,Ci);

  return(result);
}

/**Function********************************************************************

  Synopsis           [Recursively build the portioned transition relation.]

  Description        [Recursively build the portioned transition
  relation. When the size of the current cluster is greater than the
  threshold found in the options, than a new cluster is created. The
  partitions are simplified using the given <tt>assumption<tt>.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static node_ptr compileBuildBasicCPRecur(node_ptr clusters,
					 node_ptr sexp,
					 add_ptr assumption)
{
  node_ptr result;
  
  if ((sexp == Nil) || (query_CP_hash(sexp))) {
   /* if the cluster has already been considered, thus we avoid a
       further insertion */
    result = clusters;
  }
  else {
    yylineno = node_get_lineno(sexp);
    switch(node_get_type(sexp)){
    case AND:
      {
        result = compileBuildBasicCPRecur(clusters, car(sexp), assumption);
        result = compileBuildBasicCPRecur(result, cdr(sexp), assumption);
        break;
      }
    default:
      {
        add_ptr z = eval(sexp, Nil);
	CPCluster_Info_t * Ci = CPClusterInfoAlloc();

        if (add_is_one(dd_manager, z)) {
          result = clusters;
        }
        else {
          add_ptr tmp = add_simplify_assuming(dd_manager, z, assumption);

          add_free(dd_manager, z);

          Ci->Ti = add_to_bdd(dd_manager,tmp);
	  Ci->Ei = (bdd_ptr)Nil;
	  add_free(dd_manager, tmp);
          result = cons((node_ptr)Ci, clusters);
        }
        break;
      }
    }
  }
  /* We keep track that the current expression has been already
     consdered for clustering */
  insert_CP_hash(sexp);
  return(result);
}

/**Function********************************************************************

  Synopsis           [Forms the clusters of relations based on BDD
  size heuristic.]

  Description        [The clusters are formed by taking the product in
  order. Once the BDD size of the current cluster reaches a threshold, a new
  cluster is created.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void compileCPMakeClusters(DdManager * dd, node_ptr RelationList,
			   node_ptr * NewRelationList,
			   int threshold)
{
  int flag;
  node_ptr List = RelationList;
  bdd_ptr cluster, tmp_cluster, relation;

  *NewRelationList = Nil;
  while (List != Nil) {
    flag = 0;
    cluster = bdd_one(dd);
    do {
      CPCluster_Info_t * Ci = (CPCluster_Info_t *)car(List);

      relation = Ci->Ti;
      tmp_cluster = bdd_and(dd, cluster, relation);

      if ((bdd_size(dd, cluster) <= threshold && bdd_size(dd,relation) <= threshold) || (flag == 0)){
        bdd_free(dd, cluster);
        cluster = tmp_cluster;
        flag = 1;
        List = cdr(List);
      }
      else {
        bdd_free(dd, tmp_cluster);
        break;
      }
    } while (List != Nil); /* do */
    {
      CPCluster_Info_t * New_Ci = CPClusterInfoAlloc();

      New_Ci->Ti = cluster;
      *NewRelationList = cons((node_ptr)New_Ci, *NewRelationList);
    }
  } /* while (List != Nil) */
}

/**Function********************************************************************

  Synopsis           [Checks if two clusters are equal.]

  Description        [Checks if two clusters are equal. If it is the
  case then return 1 else return 0. <br>
  Notice that the check is performed only using the \"Ti\" field of the
  CPCluster_Info_t structures.] 

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int CPCLuster_equal(CPCluster_Info_t * A, CPCluster_Info_t * B)
{
  nusmv_assert(A != NULL);
  nusmv_assert(B != NULL);

  return((A->Ti == B->Ti));
}

/**Function********************************************************************

  Synopsis           [Deletes a given cluster from a list of clusters.]

  Description        [Deletes every occurrence of a given cluster from a list of clusters.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
node_ptr CPDeleteCluster(node_ptr source, CPCluster_Info_t * el)
{
  node_ptr result = Nil;
  
  /* source = reverse(source); */
  while(source != Nil) {
    if (!CPCLuster_equal((CPCluster_Info_t *)(car(source)), el))
      result = cons(car(source), result);
    source = cdr(source);
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Computes the schedule for the given list of clusters]

  Description        [Computes the schedule for the given list of clusters]

  SideEffects        [required]

  SeeAlso            [optional]

******************************************************************************/
static bdd_ptr compileBuildSchedule(node_ptr l, bdd_ptr vars)
{
  bdd_ptr supp=(bdd_ptr)Nil , tmp, result, T;
  if (l==Nil) return (bdd_ptr)bdd_one(dd_manager);
  supp = compileBuildSchedule(cdr(l),vars);
  tmp=bdd_cube_diff(dd_manager,vars,supp);
  ((CPCluster_Info_t *)car(l))->Ei = tmp;
  T=((CPCluster_Info_t *)car(l))->Ti;
  tmp = bdd_support(dd_manager,T);
  result = bdd_and(dd_manager,supp,tmp);
  bdd_free(dd_manager,tmp);
  bdd_free(dd_manager,supp);
  return(result);
}

/**Function********************************************************************

  Synopsis           [Computes the schedule for the given list of clusters]

  Description        [Computes the schedule for the given list of clusters.
                      Wrapper.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void compileCPBuildSchedule(node_ptr l, bdd_ptr vars)
{
  compileBuildSchedule(l,vars);
  return;
}

/**Function********************************************************************

  Synopsis           [Prints out the info for each cluster.]

  Description        [Prints out the information of each cluster, i.e.
  the size of the cluster in BDD nodes, the set of support of the
  cluster (distinguishing between current and next state variables),
  the variable existentially quantified when the current cluster is
  multiplied in the product. If debugging is activated then also the
  parameters necessary to compute the cost function are printed
  out.] 

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void CPPrintClustersInfo(FILE * fp, DdManager * dd,
                             node_ptr ClusterList, node_ptr var_list)
{
  node_ptr cl = ClusterList;
  int c_id = 0;
  
  while (cl != Nil) {
    CPCluster_Info_t * Ci = (CPCluster_Info_t *)car(cl);

    c_id++;
    fprintf(fp, "############################################################\n");
    fprintf(fp, "Cluster #%d:\n", c_id);
    fprintf(fp, "Size of C_%d = %d\n", c_id, bdd_size(dd, Ci->Ti));
    fprintf(fp, "\nSize of E_%d = %d Boolean Variables\n", c_id, bdd_size(dd, Ci->Ei) - 1);
    fprintf(fp, "The variables in E_%d are:\n", c_id);
    (void) print_state_vars(dd, Ci->Ei, var_list);
    fprintf(fp, "\n############################################################\n");
    cl = cdr(cl);
  }
}

/**Function********************************************************************

  Synopsis           [required]

  Description        [optional]

  SideEffects        [required]

  SeeAlso            [optional]

******************************************************************************/
void print_generic_part_info(FILE *file, Fsm_BddPtr fsm, CP_Ptr c)
{
  bdd_ptr init_bdd, invar_bdd;

  init_bdd = Compile_FsmBddGetInit(fsm);
  invar_bdd = Compile_FsmBddGetInvar(fsm);
  
  if ((init_bdd != NULL) && (invar_bdd != NULL)){
    fprintf(file, "BDD nodes representing init set of states: %d\n",
	    bdd_size(dd_manager, init_bdd));
    fprintf(file, "BDD nodes representing invar set: %d\n",
	    bdd_size(dd_manager, invar_bdd));
  }
  print_generic_part_bdd_info(file, c);
}

/**Function********************************************************************

  Synopsis           [required]

  Description        [optional]

  SideEffects        [required]

  SeeAlso            [optional]

******************************************************************************/
void print_generic_part_bdd_info(FILE *file, CP_Ptr c)
{
  int i = 0;
  node_ptr l;

  if (c != (CP_Ptr)NULL){
    fprintf(file, "Forward Partitioning Schedule BDD cluster size (#nodes):\n");
    for(l = CP_GetForwardTrans(c); l != Nil; l = cdr(l)) {
      bdd_ptr Ti = (bdd_ptr)((CPCluster_Info_t *)car(l))->Ti;
      /* fprintf(file, "BDD nodes representing partition %.3d: %d\n", ++i,
	 bdd_size(dd_manager, Ti));*/
      fprintf(file, "cluster %d\t:\t%d\n", ++i,
	      bdd_size(dd_manager, Ti));
      /*dd_printminterm(dd_manager,Ti);*/
    }
    i = 0;
    fprintf(file, "Backward Partitioning Schedule BDD cluster size (#nodes):\n");
    for(l = CP_GetBackwardTrans(c); l != Nil; l = cdr(l)) {
      bdd_ptr Ti = (bdd_ptr)((CPCluster_Info_t *)car(l))->Ti;
      /*fprintf(file, "BDD nodes representing partition %.3d: %d\n", ++i,
	      bdd_size(dd_manager, Ti));*/
      fprintf(file, "cluster %d\t:\t%d\n", ++i,
	      bdd_size(dd_manager, Ti));
      /*dd_printminterm(dd_manager, Ti);*/
    }
  }
}

/**Function********************************************************************

  Synopsis           []

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void print_generic_detailed_info(FILE *file, CP_Ptr c) {
  node_ptr Fwd_trans = CP_GetForwardTrans(c);
  node_ptr Bwd_trans = CP_GetBackwardTrans(c);
  fprintf(file, "\n########################################\n");
  fprintf(file, "# Forward clusters: \n");
  fprintf(file, "########################################\n");
  CPPrintClustersInfo(file, dd_manager,Fwd_trans, state_variables);
  fprintf(file, "\n########################################\n");
  fprintf(file, "# Backward clusters: \n");
  fprintf(file, "########################################\n");
  CPPrintClustersInfo(file, dd_manager, Bwd_trans, state_variables);
  fprintf(file, "\n");  
}

